/**
 * Copyright (C) 2007,2008 - Bull S.A.S.
 * Copyright (C) 2002-2004 - INRIA (www.inria.fr)
 *
 * CAROL: Common Architecture for RMI OW2 Layer
 *
 * This library is developed inside the OW2 Consortium,
 * http://www.ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JInterceptorStore.java 1847 2008-06-11 12:00:35Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.carol.rmi.jrmp.interceptor.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.ow2.carol.rmi.interceptor.spi.JClientRequestInterceptor;
import org.ow2.carol.rmi.interceptor.spi.JInitializer;
import org.ow2.carol.rmi.interceptor.spi.JServerRequestInterceptor;
import org.ow2.carol.rmi.jrmp.interceptor.impl.legacy.JInterceptorConverter;
import org.ow2.carol.util.configuration.TraceCarol;

/**
 * Class <code>JInterceptorStore</code> is the CAROL JRMP Client and Server
 * Interceptors Storage System
 * @author Guillaume Riviere (Guillaume.Riviere@inrialpes.fr)
 * @author Loris Bouzonnet (Add the capability to add interceptors after the Carol initialization)
 */
@SuppressWarnings("deprecation")
public final class JInterceptorStore {

    /**
     * Initializer class prefix.
     */
    public static final String INTIALIZER_PREFIX = "org.ow2.PortableInterceptor.JRMPInitializerClass";

    /**
     * Singleton.
     */
    private static JInterceptorStore jiStore = null;

    /**
     * Associations of initializer with the interceptors created by this.
     */
    private final List<JInterceptorsGroup> jinterceptorsList;

    /**
     * private remote Interceptor cache for Context propagation
     */
    private JClientRequestInterceptor[] rcis;

    /**
     * JRMP Initializer info for Carol 3.x.
     */
    private final JRMPInitInfoImpl jrmpInfo;

    /**
     * JRMP Initializer info for Carol 2.x.
     */
    private final org.objectweb.carol.rmi.jrmp.interceptor.JRMPInitInfoImpl legacyJrmpInfo;

    /**
     * Initialize interceptors for a carol server
     */
    private JInterceptorStore() {
        jrmpInfo = new JRMPInitInfoImpl();
        legacyJrmpInfo = new org.objectweb.carol.rmi.jrmp.interceptor.JRMPInitInfoImpl();
        jinterceptorsList = new ArrayList<JInterceptorsGroup>();
        // Load the Interceptors
        try {
            Properties sys = System.getProperties();
            for (Enumeration<?> e = sys.propertyNames(); e.hasMoreElements();) {
                String pkey = (String) e.nextElement();
                if (pkey.startsWith(INTIALIZER_PREFIX)) {
                    String initializer = pkey.substring(INTIALIZER_PREFIX.length() + 1);
                    Object object = getClass().getClassLoader().loadClass(initializer).newInstance();
                    JInterceptorsGroup jInterceptors;
                    if(object instanceof JInitializer) {
                        JInitializer jinit = (JInitializer) object;
                        jrmpInfo.clear();
                        jinit.preInit(jrmpInfo);
                        jinit.postInit(jrmpInfo);
                        jInterceptors =
                            new JInterceptorsGroup(initializer,
                                    jrmpInfo.getServerRequestInterceptors(),
                                    jrmpInfo.getClientRequestInterceptors());
                    } else {
                        org.objectweb.carol.rmi.jrmp.interceptor.JInitializer jinit =
                            (org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) object;
                        legacyJrmpInfo.clear();
                        jinit.pre_init(legacyJrmpInfo);
                        jinit.post_init(legacyJrmpInfo);
                        jInterceptors =
                            new JInterceptorsGroup(
                                    initializer,
                                    JInterceptorConverter.convertJServerRequestInterceptor(
                                            legacyJrmpInfo.getServerRequestInterceptors()),
                                    JInterceptorConverter.convertJClientRequestInterceptor(
                                            legacyJrmpInfo.getClientRequestInterceptors()));
                    }
                    jinterceptorsList.add(jInterceptors);
                }
            }
            // first remote reference = local reference
            resetRemoteInterceptors();
        } catch (Exception e) {
            //we did not found the interceptor do nothing but a trace ?
            TraceCarol.error("JrmpPRODelegate(), No interceptors found", e);
            throw new RuntimeException("Cannot initialize the JInterceptorStore", e);
        }
    }

    /**
     * @return the (singleton) store of JRMP interceptors
     */
    public static synchronized JInterceptorStore getJInterceptorStore() {
        if(jiStore == null) {
            jiStore = new JInterceptorStore();
        }
        return jiStore;
    }

    /**
     * Get Initializers method
     * @return JRMP Initializers enumeration
     */
    public List<JInterceptorsGroup> getJInterceptors() {
        return jinterceptorsList;
    }

    /**
     * Get interceptor if exists. Executed by the clients when unserializing a stub.
     * @param ia interceptors initializers
     * @return JClientRequestInterceptors [] , the interceptors
     */
    public synchronized JClientRequestInterceptor[] setRemoteInterceptors(final String[] ia) {
        List<JClientRequestInterceptor> result = new ArrayList<JClientRequestInterceptor>();
        for (String className : ia) {
            try {
                Object object = Class.forName(className).newInstance();
                if(object instanceof JInitializer) {
                    JInitializer jinit = (JInitializer) object;
                    jrmpInfo.clear();
                    jinit.preInit(jrmpInfo);
                    jinit.postInit(jrmpInfo);
                    result.addAll(Arrays.asList(jrmpInfo.getClientRequestInterceptors()));
                } else {
                    org.objectweb.carol.rmi.jrmp.interceptor.JInitializer jinit =
                        (org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) object;
                    legacyJrmpInfo.clear();
                    jinit.pre_init(legacyJrmpInfo);
                    jinit.post_init(legacyJrmpInfo);
                    result.addAll(Arrays.asList(JInterceptorConverter.convertJClientRequestInterceptor(
                            legacyJrmpInfo.getClientRequestInterceptors())));
                }
            } catch (Exception e) {
                TraceCarol.error("Can not load interceptors " + ia, e);
                throw new RuntimeException("Can not load interceptors " + ia, e);
            }
        }
        rcis = result.toArray(new JClientRequestInterceptor[result.size()]);
        return rcis;
    }

    /**
     * Add a new interceptor initializer.
     * @param jinitClass a class defining an interceptor initializer
     */
    public synchronized void addLocalInterceptor(final Class<?> jinitClass) {
        String jinitClassname = jinitClass.getName();
        for(JInterceptorsGroup jInterceptors : jinterceptorsList) {
            if(jInterceptors.getInitializer().equals(jinitClassname)) {
                TraceCarol.debugRmiCarol("Interceptors already registered for class " + jinitClassname);
                return;
            }
        }
        try {
            Object object = jinitClass.newInstance();
            JInterceptorsGroup jInterceptors;
            if(object instanceof JInitializer) {
                JInitializer jinit = (JInitializer) object;
                jrmpInfo.clear();
                jinit.preInit(jrmpInfo);
                jinit.postInit(jrmpInfo);
                jInterceptors =
                    new JInterceptorsGroup(jinitClassname,
                        jrmpInfo.getServerRequestInterceptors(),
                        jrmpInfo.getClientRequestInterceptors());
            } else {
                org.objectweb.carol.rmi.jrmp.interceptor.JInitializer jinit =
                    (org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) object;
                legacyJrmpInfo.clear();
                jinit.pre_init(legacyJrmpInfo);
                jinit.post_init(legacyJrmpInfo);
                jInterceptors =
                    new JInterceptorsGroup(
                            jinitClassname,
                            JInterceptorConverter.convertJServerRequestInterceptor(
                                    legacyJrmpInfo.getServerRequestInterceptors()),
                            JInterceptorConverter.convertJClientRequestInterceptor(
                                    legacyJrmpInfo.getClientRequestInterceptors()));
            }
            jinterceptorsList.add(jInterceptors);
            resetRemoteInterceptors();
        } catch (Exception e) {
            TraceCarol.error("Cannot add interceptors for class " + jinitClassname, e);
        }
    }

    /**
     * Remove an interceptor initializer.
     * @param jinitClass a class defining an interceptor initializer
     */
    public synchronized void removeLocalInterceptor(final Class<? extends JInitializer> jinitClass) {
        Iterator<JInterceptorsGroup> iterator = jinterceptorsList.iterator();
        while(iterator.hasNext()) {
            JInterceptorsGroup interceptors = iterator.next();
            if(interceptors.getInitializer().equals(jinitClass.getName())) {
                iterator.remove();
                break;
            }
        }
        resetRemoteInterceptors();
    }

    /**
     * Reset the remote interceptors.
     */
    private void resetRemoteInterceptors() {
        rcis = JInterceptorHelper.getInterceptorsByType(jinterceptorsList).getJClientRequestInterceptors();
    }

    /**
     * @return the list of initializers for JRMP interceptors
     */
    public String[] getJRMPInitializers() {
        List<String> jrmpInitializers = new ArrayList<String>();
        for(JInterceptorsGroup jInterceptorsGroup : jinterceptorsList) {
            jrmpInitializers.add(jInterceptorsGroup.getInitializer());
        }
        return jrmpInitializers.toArray(new String[jrmpInitializers.size()]);
    }

    public JClientRequestInterceptor[] getRemoteClientInterceptors() {
        return rcis;
    }

    /**
     * get the local server interceptor
     */
    public JServerRequestInterceptor[] getLocalServerInterceptors() {
        return JInterceptorHelper.getInterceptorsByType(jinterceptorsList).getJServerRequestInterceptors();
    }

    /**
     * get the local client interceptor
     */
    public JClientRequestInterceptor[] getLocalClientInterceptors() {
        return JInterceptorHelper.getInterceptorsByType(jinterceptorsList).getJClientRequestInterceptors();
    }

}
